import io
import math
import os
import sys

import numpy as np
import pandas as pd
from cctbx import crystal, miller, uctbx, xray
from cctbx.array_family import flex
from cctbx.miller import index_generator
from cctbx.sgtbx import space_group_type
from future.utils import raise_
from iotbx.cif import CifParserError, reader

from io import IOBase as file


centering_vectors = {
    'P': (['0.0', '0.0', '0.0'],),
    'A': (['0.0', '0.0', '0.0'], ['0.0', '0.5', '0.5']),
    'B': (['0.0', '0.0', '0.0'], ['0.5', '0.0', '0.5']),
    'C': (['0.0', '0.0', '0.0'], ['0.5', '0.5', '0.0']),
    'I': (['0.0', '0.0', '0.0'], ['0.5', '0.5', '0.5']),
    'R': (['0.0', '0.0', '0.0'], ['2/3', '1/3', '1/3'], ['1/3', '2/3', '2/3']),
    'S': (['0.0', '0.0', '0.0'], ['1/3', '2/3', '2/3'], ['2/3', '1/3', '1/3']),
    'T': (['0.0', '0.0', '0.0'], ['1/3', '2/3', '2/3'], ['2/3', '1/3', '2/3']),
    'H': (['0.0', '0.0', '0.0'], ['2/3', '1/3', '0.0'], ['1/3', '2/3', '0.0']),
    'K': (['0.0', '0.0', '0.0'], ['1/3', '0.0', '2/3'], ['2/3', '0.0', '1/3']),
    'L': (['0.0', '0.0', '0.0'], ['0.0', '2/3', '1/3'], ['0.0', '1/3', '2/3']),
    'F': (['0.0', '0.0', '0.0'], ['0.0', '0.5', '0.5'], ['0.5', '0.0', '0.5'], ['0.5', '0.5', '0.0'])
}


def read_cif(f):
    """opens cif and returns cctbx data object"""
    try:
        if isinstance(f, io.IOBase):
            structures = reader(file_object=f).build_crystal_structures()
        elif isinstance(f, str):
            structures = reader(file_path=f).build_crystal_structures()
        else:
            raise TypeError(f'read_cif: Can not deal with type {type(f)}')
    except CifParserError as e:
        print(e)
        print("Error parsing cif file, check if the data tag does not contain any spaces.")
        sys.exit()
    for key, val in list(structures.items()):
        print("\nstructure:", key)
        val.show_summary().show_scatterers()
    return structures


def load_hkl(fin, labels=None, shelx=False, savenpy=False, verbose=True):
    """Read a file with filename 'fin', labels describe the data in the columns.
    The h,k,l columns are labeled by default and expected to be the first 3 columns.
    Returns a dataframe with hkl values as the indices

    e.g. load_hkl('red.hkl',labels=('F','sigmas')

    All columns should be labeled and columns labeled:
        None,'None','none' or 'skip' will be ignored

    It is recommended to label all expected columns, so the algorithm will return an
    error rather than read the next column when columns are not delimited

    savenpy: save data as binary numpy format for faster loading on next use."""

    if shelx == True and labels == None:
        labels = ('F2', 'sigma')
    elif labels == None:
        raise TypeError('load_hkl() did not get a value for labels.')

    if isinstance(fin, file):
        fname = fin.name
    else:
        fname = fin

    labels = tuple(labels)
    skipcols = (None, 'None', 'none', 'skip') + \
        tuple(item for item in labels if item.startswith('#'))
    usecols = [0, 1, 2] + \
        [3+i for i, label in enumerate(labels) if label not in skipcols]

    root, ext = os.path.splitext(fname)
    changed = False

    try:
        inp = np.load(root+'.npy')
        assert len(inp.T) == len(
            usecols), 'npy data did not match to expected columns'
    except (OSError, AssertionError):
        changed = True
        if shelx == False:
            try:
                # if this fails, try shelx file
                inp = np.loadtxt(fin, usecols=usecols)
            except ValueError:
                inp = np.genfromtxt(
                    fin, usecols=usecols, delimiter=[4, 4, 4, 8, 8, 4])
        else:
            inp = np.genfromtxt(
                fin, usecols=usecols, delimiter=[4, 4, 4, 8, 8, 4])
    else:
        ext = '.npy'
        fname = root+'.npy'

    if savenpy:
        if ext != '.npy' or changed:
            print(f'Writing data as npy format to {fname}')
            np.save(root, inp)

    if verbose:
        print('')
        print(f'Loading data: {fname}')
        print(f'     usecols: {usecols}')
        print('      labels: {}'.format(' '.join(('h', 'k', 'l')+labels)))
        print(f'       shape: {inp.shape}')
    else:
        print(f'Loading data: {fname} => ({inp.shape[0]:5d}, {inp.shape[1]:2d})')

    h = list(map(int, inp[:, 0]))
    k = list(map(int, inp[:, 1]))
    l = list(map(int, inp[:, 2]))

    index = list(zip(h, k, l))

    labels = (label for label in labels if label not in skipcols)

    d = dict(list(zip(labels, inp[:, 3:].T)))

    df = pd.DataFrame(d, index=index)

    if not df.index.is_unique:
        print(f"\n** Warning: Duplicate indices detected in {fname} **\n")

        # useful, but very slow for large data sets!
        # index = list(df.index)
        # duplicates = set([x for x in index if index.count(x) > 1])
        # for x in duplicates: print x

    return df


def f_calc_dspacing(a, b, c, al, be, ga, h, k, l, kind='triclinic'):
    """
    Calculates d-spacing based on given parameters.
    a,b,c,al,be,ge are given as floats
    al,be,ga can be given as ndarrays or floats
    kind specifies the type of cell -> triclinic works for the general case, but is a bit slower
    although still fast enough

    Tested: orthorhombic cell on (orthorhombic, monoclinic, triclinic)
    Tested: triclinic cell with dvalues from topas
    """

    if kind == 'cubic':
        print('\n** Warning: cubic dspacing calculation unverified!! **\n')
        idsq = (h**2 + k**2 + l**2) / a**2

    elif kind == 'tetragonal':
        print('\n** Warning: tetragonal dspacing calculation unverified!! **\n')
        idsq = (h**2 + k**2) / a**2 + l**2 / c**2

    elif kind == 'orthorhombic':
        idsq = h**2 / a**2 + k**2 / b**2 + l**2 / c**2

    elif kind == 'hexagonal':
        print('\n** Warning: hexagonal dspacing calculation unverified!! **\n')
        idsq = (4/3) * (h**2 + h*k + k**2) * (1/a**2) + l**2 / c**2

    elif kind == 'monoclinic':
        print('\n** Warning: monoclinic dspacing calculation unverified!! **\n')
        be = math.radians(be)
        idsq = (1/math.sin(be)**2) * (h**2/a**2 + k**2 * math.sin(be)
                                      ** 2 / b**2 + l**2/c**2 - (2*h*l*math.cos(be)) / (a*c))

    elif kind == 'triclinic':
        V = calc_volume(a, b, c, al, be, ga)

        al = math.radians(al)
        be = math.radians(be)
        ga = math.radians(ga)

        idsq = (1/V**2) * (
            h**2 * b**2 * c**2 * math.sin(al)**2
            + k**2 * a**2 * c**2 * math.sin(be)**2
            + l**2 * a**2 * b**2 * math.sin(ga)**2
            + 2*h*k*a*b*c**2 * (math.cos(al) * math.cos(be) - math.cos(ga))
            + 2*k*l*b*c*a**2 * (math.cos(be) * math.cos(ga) - math.cos(al))
            + 2*h*l*c*a*b**2 * (math.cos(al) * math.cos(ga) - math.cos(be))
        )

    d = 1/idsq**0.5

    return d


def calc_dspacing(df, cell, col='d', kind='triclinic', inplace=True):
    """Calculate dspacing on df from indices"""
    a, b, c, al, be, ga = cell
    h, k, l = list(map(np.array, list(zip(*df.index))))
    d = f_calc_dspacing(a, b, c, al, be, ga, h, k, l, kind=kind)
    if inplace:
        df[col] = d
    else:
        return d


def calc_volume(a, b, c, al, be, ga):
    """Returns volume for the general case from cell parameters"""
    al = math.radians(al)
    be = math.radians(be)
    ga = math.radians(ga)
    V = a*b*c*((1+2*math.cos(al)*math.cos(be)*math.cos(ga) -
                math.cos(al)**2-math.cos(be)**2-math.cos(ga)**2)**.5)
    return V


def f_calc_structure_factors(structure, **kwargs):
    """Takes cctbx structure and returns f_calc miller array
    Takes an optional options dictionary with keys:
    input:
        **kwargs:
            'd_min': minimum d-spacing for structure factor calculation
            'algorithm': which algorithm to use ('direct', 'fft', 'automatic')
        structure: <cctbx.xray.structure.structure object>
    output:
        f_calc: <cctbx.miller.array object> with calculated structure factors
            in the f_calc.data() function

    TODO:
    - make this more general?
    - allow for specification of more parameters (like tables, ie. it1992 or wk1995)
    """

    dmin = kwargs.get('dmin', 1.0)
    algorithm = kwargs.get('algorithm', "automatic")
    anomalous = kwargs.get('anomalous', False)
    table = kwargs.get('scatfact_table', 'wk1995')
    return_as = kwargs.get('return_as', "series")

    if dmin <= 0.0:
        raise ValueError("d-spacing must be greater than zero.")

    if algorithm == "automatic":
        if structure.scatterers().size() <= 100:
            algorithm = "direct"
        else:
            algorithm = None

    structure.scattering_type_registry(table=table)

    f_calc_manager = structure.structure_factors(
        anomalous_flag=anomalous,
        d_min=dmin,
        algorithm=algorithm)
    f_calc = f_calc_manager.f_calc()

    print("\nScattering table:", structure.scattering_type_registry_params.table)
    structure.scattering_type_registry().show()
    print("Minimum d-spacing: %g" % f_calc.d_min())

    if return_as == "miller":
        return f_calc
    elif return_as == "series":
        index = list(f_calc.indices())
        fcalc = pd.Series(index=index, data=np.abs(f_calc.data()))
        phase = pd.Series(index=index, data=np.angle(f_calc.data()))
        return fcalc, phase
    elif return_as == "df":
        dffcal = pd.DataFrame(index=f_calc.index)
        dffcal['fcalc'] = np.abs(f_calc.data())
        dffcal['phase'] = np.angle(f_calc.data())
        return dffcal
    else:
        raise_(ValueError, f"Unknown argument for 'return_as':{return_as}")


def calc_structure_factors(cif, dmin=1.0, combine=None, table='xray', prefix='', **kwargs):
    """Wrapper around f_calc_structure_factors()
    Takes a cif file (str or file object)

    dmin can be a dataframe and it will take the minimum dspacing (as specified by col 'd') or a float
    if combine is specified, function will return a dataframe combined with the given one, otherwise a
    dictionary of dataframes

    prefix is a prefix for the default names fcalc/phases to identify different structures"""

    if isinstance(cif, str):
        f = open(cif)

    if isinstance(dmin, pd.DataFrame):
        dmin = min(dmin['d']) - 0.00000001

    structures = read_cif(f)

    if isinstance(structures, xray.structure):
        structures = {"fcalc": structures}

    col_phases = prefix+"phases"
    col_fcalc = prefix+"fcalc"

    for name, structure in list(structures.items()):
        fcalc, phase = f_calc_structure_factors(
            structure, dmin=dmin, scatfact_table=table, return_as="series", **kwargs)

        if len(structures) > 1:
            col_phases = prefix+"ph_"+name
            col_fcalc = prefix+"fc_"+name

        dffcal = pd.DataFrame({col_phases: phase, col_fcalc: fcalc})

        if combine is not None:
            combine = combine.combine_first(dffcal)
        else:
            structures[name] = dffcal

    if combine is not None:
        try:
            return combine.sort('d', ascending=False)
        except KeyError:
            return combine
    else:
        return structures


def merge_sym_equiv(m, output='ma', algorithm=None, verbose=True):
    """takes miller.array, returns hkl dictionary
    http://cci.lbl.gov/cctbx_sources/cctbx/miller/equivalent_reflection_merging.tex

    testing merging algorithms on different data sets:
    works = tested to give expected results
    fails = tested not to give expected results

                       gaussian        shelx
    unique             works           works
    unique+sigmas
    not unique                         works
    not unique+sigmas

    With no sigmas, shelx takes average of present refs in each group
    """
    n_unique = len(m.unique_under_symmetry().data())
    n_refs = len(m.data())

    # The number of merged reflections may be more than expected for powder data
    # this is because sometimes multiple indexes are read into the dataframe for some data sets
    # merging should then complete with R = 0, which is ok for powders

    if not algorithm:
        # indicates powder data set? => see also note above
        if n_refs == n_unique:
            # works with unique data, but gives strange results for single
            # crystal data
            algorithm = 'shelx'
        else:
            # gaussian returns 0 for unique data set, such as powders
            algorithm = 'shelx'

    # print "\ntotal/unique = {}/{} -> merging algorithm == {}".format(n_refs,n_unique,algorithm)
    # print "Using gaussian NEEDS accurate sigmas, shelx doesn't."

    merging = m.merge_equivalents(algorithm=algorithm)

    if verbose:
        print()
        merging.show_summary()
        print()

    m_out = merging.array()
    print('{} reflections merged/averaged to {}'.format(m.size(), m_out.size()))
    if output == 'dict':
        return miller_array_to_dict(m_out)
    elif output == 'ma':
        return m_out


def remove_sysabs(m, verbose=True):
    """Returns new miller.array with systematic absences removed"""
    sysabs = m.select_sys_absent().sort('data')

    if sysabs.size() > 0 and verbose:
        print(f'\nTop 10 systematic absences removed (total={sysabs.size()}):')

        if sysabs.sigmas() == None:
            for ((h, k, l), sf) in sysabs[0:10]:
                print(f'{h:4}{k:4}{l:4} {sf:8.2f}')
        else:
            for ((h, k, l), sf, sig) in sysabs[0:10]:
                print(f'{h:4}{k:4}{l:4} {sf:8.2f} {sig:8.2f}')

        print("Compared to largest 3 reflections:")
        if m.sigmas() == None:
            for ((h, k, l), sf) in m.sort('data')[0:3]:
                print(f'{h:4}{k:4}{l:4} {sf:8.2f}')
        else:
            for ((h, k, l), sf, sig) in m.sort('data')[0:3]:
                print(f'{h:4}{k:4}{l:4} {sf:8.2f} {sig:8.2f}')

        return m.remove_systematic_absences()
    elif sysabs.size() > 0:
        print(f"{sysabs.size()} systematic absences removed")
        return m.remove_systematic_absences()
    else:
        return m


def f_calc_multiplicities(df, cell, spgr):
    """Small function to calculate multiplicities for given dataframe"""
    m = df2m(df, cell, spgr)
    df['m'] = list(m.multiplicities().data())

def make_symmetry(cell, spgr):
    """takes cell parameters (a,b,c,A,B,C) and spacegroup (str, eg. 'cmcm'), returns cctbx
    crystal_symmetry class required for merging of reflections"""
    if not cell:
        cell = input("Please specify a cell:\n >> ")
        cell = list(map(float, cell.split()))
    if not spgr:
        spgr = input("Please specify space group:\n >> ")

    crystal_symmetry = crystal.symmetry(
        unit_cell=cell,
        space_group_symbol=spgr)
    return crystal_symmetry


def generate_indices(cell, spgr, dmin=1.0, ret='index'):
    """http://cci.lbl.gov/cctbx_sources/cctbx/miller/index_generator.h"""

    dmin = dmin-0.0000000000001  # because resolution_limit is < and not <=

    anomalous_flag = False
    symm = make_symmetry(cell, spgr)

    unit_cell = uctbx.unit_cell(cell)
    sg_type = space_group_type(spgr)
    # input hkl or resolution(d_min)
    mig = index_generator(
        unit_cell, sg_type, anomalous_flag=anomalous_flag, resolution_limit=dmin)
    indices = mig.to_array()
    if ret == 'index':  # suitable for df index
        return list(indices)
    else:
        return list(miller.array(miller.set(crystal_symmetry=symm, indices=indices, anomalous_flag=anomalous_flag)))


def m2df(m, data='data', sigmas='sigmas'):
    """Takes a miller.array object and returns the
    m: miller.array
    data and sigmas are the names for the columns in the resulting dataframe
    if no data/sigmas are present in the miller array, these are ignored.
    """
    index = list(m.indices())
    df = pd.DataFrame(index=index)

    if m.data():
        df[data] = list(m.data())
    if m.sigmas():
        df[sigmas] = list(m.sigmas())

    return df


def df2m(df, cell, spgr, data=None, sigmas=None):
    """Constructs a miller.array from the columns specified by data/sigmas in the dataframe,
    if both are None, returns just the indices.
    needs cell and spgr to generate a symmetry object."""
    anomalous_flag = False

    if isinstance(df, pd.DataFrame):
        if data:
            try:
                sel = df[data].notnull()  # select notnull data items for index
            except ValueError:
                index = df.index
            else:
                index = df.index[sel]
        else:
            index = df.index
    else:
        index = df

    indices = flex.miller_index(index)

    if data:
        data = flex.double(df[data][sel])
    if sigmas:
        sigmas = flex.double(df[sigmas][sel])

    symm = make_symmetry(cell, spgr)
    ms = miller.set(
        crystal_symmetry=symm, indices=indices, anomalous_flag=anomalous_flag)
    return miller.array(ms, data=data, sigmas=sigmas)


def reduce_all(df, cell, spgr, dmin=None, reindex=True, verbose=True):
    """Should be run after files2df. Takes care of some things that have to be done anyway.
    Once data has been loaded, this function reduces and merges all the data to a single
    unique set, adds dspacings and multiplicities and orders columns and sorts by the dspacing.

    dmin:
    can be the name of a column and dmin is taken from that
    can be float
    can be None and the dmin is determined automatically asl the largest dmin of all columns read

    reindex: bool
    will reindex using all indices generated up to dmin

    Returns a dataframe object"""

    d = pd.Series(calc_dspacing(df, cell, inplace=False), index=df.index)
    dmins = [d[df[col].notnull()].min() for col in df if col not in ('m', 'd')]

    # little table with dmins
    cols = [col for col in df if col not in ('m', 'd')]
    ln = max([len(col) for col in cols])
    print('\n{:>{}} {:>6s}'.format('', ln, 'dmin'))
    for col, dval in zip(cols, dmins):
        print('{:>{}} {:6.3f}'.format(col, ln, dval))

    if not dmin:
        # find the largest dmin for all data sets
        dmin = max(dmins)
    elif isinstance(dmin, str):
        sel = df[dmin].notnull()
        dmin = min(d[sel])
    else:
        dmin = float(dmin)

    order = ['m', 'd']
    order = order + [col for col in df if col not in order]

    dfm = pd.DataFrame()

    for col in df:
        if col in ('m', 'd'):
            continue
        print(f'\n - Merging {col}: ')
        m = df2m(df, cell=cell, spgr=spgr, data=col)
        m = remove_sysabs(m, verbose=verbose)
        m = merge_sym_equiv(m, verbose=verbose)
        dfm = dfm.combine_first(m2df(m, data=col))

    if reindex:
        index = generate_indices(cell, spgr, dmin)
        index = pd.Index(index)
        dfm = dfm.reindex(index=index)
        dfm = dfm.reindex(columns=order)

    f_calc_multiplicities(dfm, cell, spgr)
    calc_dspacing(dfm, cell)

    print(f"\nReduced/merged data to dmin = {dmin}, {len(dfm)} refs")

    return dfm.sort_values(by='d', ascending=False)


def write_hkl(df, cols=None, out=None, no_hkl=False, pre=None, post=None, data_fmt=None, hkl_fmt=None):
    """Function for writing indices + selected columns to specified file/file object or terminal."""

    if isinstance(pre, list):
        if all('\n' in line for line in pre):
            pre = ''.join(pre)
        else:
            pre = '\n'.join(pre)
    elif isinstance(pre, str):
        pre = '\n'.join(pre.strip('\n'))

    if isinstance(post, list):
        post = ''.join(post)

    if not cols:
        cols = df.columns

    if isinstance(cols, str):
        cols = (cols,)

    if isinstance(out, str):
        out = open(out, 'w')

    cols = list(cols)

    if not hkl_fmt:
        if no_hkl:
            hkl_fmt = ''
        else:
            hkl_fmt = '{:4d}{:4d}{:4d}'

    if not data_fmt:
        ifmt = '{:4d}'
        dfmt = ' {:5d}'
        ffmt = ' {:9.3f}'
        bfmt = ' {:4}'

        n = len(cols)
        data_fmt = ''

        for item in cols[:]:
            if item == '*':
                cols.remove('*')
                data_fmt += '  *  '
                continue

            # tp = repr(type(df[item][0]))
            tp = repr(df[item].dtype)
            if 'int' in tp:
                data_fmt += dfmt
            elif 'float' in tp:
                data_fmt += ffmt
            elif 'bool' in tp:
                data_fmt += bfmt
            else:
                raise TypeError(f"No format associated with type {tp}")
    elif data_fmt == 'shelx':
        data_fmt = '{:8.3f}{:8.3f}'

    if pre:
        print(pre, file=out)

    print('>> Writing {} refs to file {}'.format(len(df), out.name if out else 'stdout'))

    for row in df.reindex(columns=cols).itertuples():

        # if (abs(row[1:][2] - row[1:][4]) < 0.0001 and
        #     abs(row[1:][2] - row[1:][5]) < 0.0001 and
        #     abs(row[1:][2] - row[1:][6]) < 0.0001):
        #   continue

        print(hkl_fmt.format(*row[0])+data_fmt.format(*row[1:]), file=out)

    if post:
        print(post, file=out)



import io
import sys

from iotbx.cif import CifParserError, reader


def read_cif(f, verbose=False):
    """opens cif and returns cctbx data object"""
    try:
        if isinstance(f, io.IOBase):
            structures = reader(file_object=f).build_crystal_structures()
        elif isinstance(f, str):
            structures = reader(file_path=f).build_crystal_structures()
        else:
            raise TypeError(
                f'read_cif: Can not deal with type {type(f)}')
    except CifParserError as e:
        print(e)
        print("Error parsing cif file, check if the data tag does not contain any spaces.")
        sys.exit()
    if verbose:
        for key, val in list(structures.items()):
            print("\nstructure:", key)
            val.show_summary().show_scatterers()
    return structures


def main():
    cifs = sys.argv[1:]

    if len(cifs) == 0:
        print("cif2topas - tool for converting cif files to Topas format (.inp)")
        print("Takes any number of cif files and writes to stdout")
        print("")
        print("usage: cif2topas cif1 [cif2 ...]")
        sys.exit()

    for cif in cifs:
        # print 'Reading file: {}'.format(cif)
        structures = read_cif(cif)
        for name in structures:
            structure = structures[name].sites_mod_positive()
            sg = structure.space_group()
            uc = structure.unit_cell()
            sps = structure.special_position_settings()
            scatterers = structure.scatterers()

            print("""
   str
      Out_X_Yobs(x_yobs.xy)
      Out_X_Ycalc(x_ycalc.xy)
      Out_X_Difference(x_ydiff.xy)
      Create_2Th_Ip_file(ticks.out)

      Out_fobs(fobs.hkl)

      Out_CIF_STR(structure.cif)""")

            print("""       
      scale  @  1.0
      r_bragg  1.0""")
            print()

            s400 = "      'prm s400 1.0 \n"
            s040 = "      'prm s040 1.0 \n"
            s004 = "      'prm s004 1.0 \n"
            s220 = "      'prm s220 1.0 \n"
            s202 = "      'prm s202 1.0 \n"
            s022 = "      'prm s022 1.0 \n"
            s301 = "      'prm s301 1.0 \n"
            s121 = "      'prm s121 1.0 \n"
            s103 = "      'prm s103 1.0 \n"
            eta = "      'prm !eta 0.5 min 0.0 max 1.0\n"
            if sg.crystal_system() == 'Monoclinic':
                string = "{s400}{s040}{s004}{s220}{s202}{s022}{s301}{s121}{s103}{eta}".format(
                    s400=s400, s040=s040, s004=s004, s220=s220, s202=s202, s022=s022, s301=s301, s121=s121, s103=s103, eta=eta)
                macro = "      'Stephens_monoclinic(s400, s040, s004, s220, s202, s022, s301, s121, s103, eta)"

            elif sg.crystal_system() == 'Tetragonal':
                string = "{s400}{s004}{s220}{s202}{eta}".format(
                    s400=s400, s040=s040, s004=s004, s220=s220, s202=s202, s022=s022, s301=s301, s121=s121, s103=s103, eta=eta)
                macro = "      'Stephens_tetragonal(s400, s004, s220, s202, eta)"

            elif sg.crystal_system() == 'Hexagonal':
                string = "{s400}{s004}{s202}{eta}".format(
                    s400=s400, s040=s040, s004=s004, s220=s220, s202=s202, s022=s022, s301=s301, s121=s121, s103=s103, eta=eta)
                macro = "      'Stephens_hexagonal(s400, s202, s004, eta)"

            elif sg.crystal_system() == 'Orthorhombic':
                string = "{s400}{s040}{s004}{s220}{s202}{s022}{eta}".format(
                    s400=s400, s040=s040, s004=s004, s220=s220, s202=s202, s022=s022, s301=s301, s121=s121, s103=s103, eta=eta)
                macro = "      'Stephens_orthorhombic(s400, s040, s004, s220, s202, s022, eta)"
            else:
                string = ""
                macro = ""
            print(string, end=' ')
            print(macro)

            print("""
      PV_Peak_Type(
      ha,    0.02,
      !hb,   0.0,
      !hc,   0.0,
      lora,  0.5,
      !lorb, 0.0,
      !lorc, 0.0)

      'Simple_Axial_Model(axial, 0.0)

      Phase_Density_g_on_cm3( 1.0)""")

            print("""
      view_structure

      'fourier_map 1
      '   fourier_map_formula = Fobs - Fcalc;""")
            print()
            print('      space_group "{}"'.format(
                sg.type().universal_hermann_mauguin_symbol().replace(" ", "")))
            print()
            a, b, c, al, be, ga = uc.parameters()
            if sg.crystal_system() == 'Cubic':
                string = "      Cubic(@ {a})"

            elif sg.crystal_system() == 'Tetragonal':
                string = "      Tetragonal(@ {a}, @ {c})"

            elif sg.crystal_system() == 'Hexagonal':
                string = "      Hexagonal(@ {a}, @ {c})"

            elif sg.crystal_system() == 'Rhombohedral':
                string = "      Rhombohedral(@ {a}, @ {al})"
            else:
                string = """      a  @  {a:8.5f}
      b  @  {b:8.5f}
      c  @  {c:8.5f}
      al {refal:1}  {al}
      be {refbe:1}  {be}
      ga {refga:1}  {ga}"""
            refal = "" if al == 90 else "@"
            refbe = "" if be == 90 else "@"
            refga = "" if ga == 90 else "@"
            print(string.format(a=a, b=b, c=c, refal=refal,
                  al=al, refbe=refbe, be=be, refga=refga, ga=ga))
            print()
            print(f"      volume {uc.volume():.2f}")
            print()
            for element in set(scatterers.extract_scattering_types()):
                print(f"      prm beq{element:2s}  2.0  min 1.0  max 5.0")
            print()

            z_order = sg.order_z()

            for atom in scatterers:
                label = atom.label
                x, y, z = atom.site
                atom.element_symbol()
                element = atom.element_symbol()
                mult = atom.multiplicity()

                print("      site {label:5s}  num_posns {mult:3d}  x  {x:.5f}  y  {y:.5f}  z  {z:.5f}  occ {element:2s}  1.0  beq =beq{element:2s};".format(
                    label=label, mult=mult, x=x, y=y, z=z, element=element), end=' ')

                if mult < z_order:
                    print("   ' {:5s} {:5s} {:5s}".format(
                        *sps.site_symmetry(atom.site).special_op_simplified().terms))
                else:
                    print()


if __name__ == '__main__':
    main()



from cctbx import crystal

centering_vectors = {
    'P': (['0.0', '0.0', '0.0'],),
    'A': (['0.0', '0.0', '0.0'], ['0.0', '0.5', '0.5']),
    'B': (['0.0', '0.0', '0.0'], ['0.5', '0.0', '0.5']),
    'C': (['0.0', '0.0', '0.0'], ['0.5', '0.5', '0.0']),
    'I': (['0.0', '0.0', '0.0'], ['0.5', '0.5', '0.5']),
    'R': (['0.0', '0.0', '0.0'], ['2/3', '1/3', '1/3'], ['1/3', '2/3', '2/3']),
    'S': (['0.0', '0.0', '0.0'], ['1/3', '2/3', '2/3'], ['2/3', '1/3', '1/3']),
    'T': (['0.0', '0.0', '0.0'], ['1/3', '2/3', '2/3'], ['2/3', '1/3', '2/3']),
    'H': (['0.0', '0.0', '0.0'], ['2/3', '1/3', '0.0'], ['1/3', '2/3', '0.0']),
    'K': (['0.0', '0.0', '0.0'], ['1/3', '0.0', '2/3'], ['2/3', '0.0', '1/3']),
    'L': (['0.0', '0.0', '0.0'], ['0.0', '2/3', '1/3'], ['0.0', '1/3', '2/3']),
    'F': (['0.0', '0.0', '0.0'], ['0.0', '0.5', '0.5'], ['0.5', '0.0', '0.5'], ['0.5', '0.5', '0.0'])
}


def make_special_position_settings(cell, space_group, min_dist_sym_equiv=0.000001):
    """Takes cell and space group, returns cctbx structure
    input:  cell: (a b c alpha beta gamma) as a tuple
            space_group: 'space_group' like a string
            min_dist_sym_equiv: float
    output: <cctbx.crystal.special_position_settings object>
                contains cell, space group and special positions
    """

    assert type(cell) == tuple, 'cell must be supplied as a tuple'
    assert len(cell) == 6, 'expected 6 cell parameters'

    special_position_settings = crystal.special_position_settings(
        crystal_symmetry=crystal.symmetry(
            unit_cell=cell,
            space_group_symbol=space_group),
        min_distance_sym_equiv=min_dist_sym_equiv)
    return special_position_settings


def print_superflip(fcalc, fout, fdiff_file=None):
    """Prints an inflip file that can directly be used with superflip for difference fourier maps

    - Tested and works fine with: EDI, SOD
    """
    print('title', 'superflip\n', file=fout)

    print('dimension 3', file=fout)
    print('voxel', end=' ', file=fout)
    for p in fcalc.unit_cell().parameters()[0:3]:
        print(int(((p*4) // 6 + 1) * 6), end=' ', file=fout)
    print(file=fout)
    print('cell', end=' ', file=fout)
    for p in fcalc.unit_cell().parameters():
        print(p, end=' ', file=fout)
    print('\n', file=fout)

    print('centers', file=fout)
    for cvec in centering_vectors[fcalc.space_group_info().type().group().conventional_centring_type_symbol()]:
        print(' '.join(cvec), file=fout)
    print('endcenters\n', file=fout)

    print('symmetry #', fcalc.space_group_info().symbol_and_number(), file=fout)
    print('# inverse no', file=fout)

    # number of unique symops, no inverses
    n_smx = fcalc.space_group_info().type().group().n_smx()
    # number of primitive symops, includes inverses
    order_p = fcalc.space_group_info().type().group().order_p()
    # total number of symops
    order_z = fcalc.space_group_info().type().group().order_z()

    # this should work going by the assumption that the unique primitive symops are stored first,
    # THEN the inverse symops and then all the symops due to centering.

    for n, symop in enumerate(fcalc.space_group_info().type().group()):
        if n == order_p:
            break
        elif n == n_smx:
            print('# inverse yes, please check!', file=fout)
        print(symop, file=fout)

        # Broken, because .inverse() doesn't work, but probably a better approach:
    # for symop in f.space_group_info().type().group().smx():
    #   print >> fout, symop
    # if f.space_group_info().type().group().is_centric():
    #   print >> fout, '# inverse yes'
    #   for symop in f.space_group_info().type().group().smx():
    #       print >> fout, symop.inverse() # inverse does not work?

    print('endsymmetry\n', file=fout)

    print('perform fourier', file=fout)
    print('terminal yes\n', file=fout)

    print('expandedlog yes', file=fout)
    print('outputfile superflip.xplor', file=fout)
    print('outputformat xplor\n', file=fout)

    print('dataformat amplitude phase', file=fout)

    if fdiff_file:
        print('fbegin fdiff.out\n', file=fout)
    else:
        print('fbegin', file=fout)
        print_simple(fcalc, fout, output_phases='cycles')

#       for i,(h,k,l) in enumerate(f.indices()):
#           # structurefactor = abs(f.data()[i])
#           # phase = phase(f.data()[i]
#           print >> fout, "%3d %3d %3d %10.6f %10.3f" % (
#               h,k,l, abs(f.data()[i]), phase(f.data()[i]) / (2*pi) )
        print('endf', file=fout)


def make_superflip(cell, spgr, wavelength, composition, datafile, dataformat, filename='sf.inflip'):
    sps = make_special_position_settings(cell, spgr)
    sg = sps.space_group()
    uc = sps.unit_cell()
    # sgi = sps.space_group_info()

    fout = open(filename, 'w')

    print('title', filename.split('.')[0], file=fout)
    print(file=fout)
    print('dimension 3', file=fout)
    print('voxel', end=' ', file=fout)
    for p in uc.parameters()[0:3]:
        print(int(((p*4) // 6 + 1) * 6), end=' ', file=fout)
    print(file=fout)
    print('cell', end=' ', file=fout)
    for p in uc.parameters():
        print(p, end=' ', file=fout)
    print(f'  # vol = {uc.volume():.4f} A3 \n', file=fout)

    print('centers', file=fout)
    for cvec in centering_vectors[sg.conventional_centring_type_symbol()]:
        print('  ', ' '.join(cvec), file=fout)
    print('endcenters\n', file=fout)

    print('symmetry #', sg.crystal_system(), sg.info(), file=fout)
    print('# +(0 0 0) Inversion-Flag = 0', file=fout)

    n_smx = sg.n_smx()
    order_p = sg.order_p()
    order_z = sg.order_z()

    for n, symop in enumerate(sg):
        if n == order_p:
            break
        elif n == n_smx:
            print('# +(0 0 0) Inversion-Flag = 1', file=fout)
        print('  ', symop, file=fout)
    print('endsymmetry\n', file=fout)
    print('derivesymmetry yes', file=fout)
    print('searchsymmetry average', file=fout)
    print(file=fout)
    print('delta AUTO', file=fout)
    print('weakratio 0.00', file=fout)
    print('biso 2.0', file=fout)
    print('randomseed AUTO', file=fout)
    print(file=fout)
    if composition:
        print(f'composition {composition}', file=fout)
        print('histogram composition', file=fout)
        print('hmparameters 10 5', file=fout)
    else:
        print('#composition #composition goes here', file=fout)
        print('#histogram composition', file=fout)
        print('#hmparameters 10 5', file=fout)
    print(file=fout)
    print('fwhmseparation 0.3', file=fout)
    print(f'lambda {wavelength}', file=fout)
    print(file=fout)
    print('maxcycles 200', file=fout)
    print('bestdensities 10', file=fout)
    print('repeatmode 100', file=fout)
    print(file=fout)
    print('polish yes', file=fout)
    print('convergencemode never', file=fout)
    print(file=fout)
    print('#referencefile filename.cif', file=fout)
    print('#modelfile filename.cif 0.2', file=fout)
    print(file=fout)
    print('terminal yes', file=fout)
    print('expandedlog yes', file=fout)
    outputfile = str(sg.info()).replace(' ', '').replace('/', 'o').lower()
    print(f'outputfile {outputfile}.xplor {outputfile}.ccp4', file=fout)
    print('outputformat xplor ccp4', file=fout)
    print(file=fout)
    print('dataformat', dataformat, file=fout)
    print(f'fbegin {datafile}\n', file=fout)


def main(filename='sf.inflip'):
    """Creates a basic superflip input file for structure solution by asking a few simple questions"""

    for x in range(3):
        cell = input("Enter cell parameters:\n >> ")

        cell = cell.split()
        if len(cell) != 6:
            print('Expecting 6 parameters: a b c alpha beta gamma')
            continue
        else:
            try:
                cell = tuple(map(float, cell))
            except ValueError as e:
                print('ValueError:', e)
                continue
            else:
                break

    for x in range(3):
        spgr = input('Enter space group:\n >> ')

        if not spgr.split():
            continue
        else:
            break

    wavelength = input('Enter wavelength\n >> [1.54056] ') or '1.54056'
    composition = input('Enter composition:\n >> [skip] ') or ''
    datafile = input('Enter datafile:\n >> [fobs.out] ') or 'fobs.out'

    for x in range(3):
        dataformat = input(
            'Enter dataformat:\n >> [intensity fwhm] ') or 'intensity fwhm'
        if not all(i in ('intensity', 'amplitude', 'amplitude difference', 'a', 'b', 'phase', 'group', 'dummy', 'fwhm', 'm91', 'm90', 'shelx')
                   for i in dataformat.split()):
            print('Unknown dataformat, please enter any of\n intensity/amplitude/amplitude difference/a/b/phase/group/dummy/fwhm/m91/m90/shelx\n')
            continue
        else:
            break

    make_superflip(
        cell, spgr, wavelength, composition, datafile, dataformat, filename=filename)


if __name__ == '__main__':
    main()



import argparse
import io
import os
import sys

from cctbx import crystal, xray
from cctbx.array_family import flex
from iotbx.cif import CifParserError, reader


def read_cif(f):
    """opens cif and returns cctbx data object"""
    try:
        if isinstance(f, io.IOBase):
            structures = reader(file_object=f).build_crystal_structures()
        elif isinstance(f, str):
            structures = reader(file_path=f).build_crystal_structures()
        else:
            raise TypeError(f'read_cif: Can not deal with type {type(f)}')
    except CifParserError as e:
        print(e)
        print("Error parsing cif file, check if the data tag does not contain any spaces.")
        sys.exit()
    for key, val in list(structures.items()):
        print("\nstructure:", key)
        val.show_summary().show_scatterers()
    return structures


def main():
    description = "expandcell is a tool to create supercells from a cif file."

    parser = argparse.ArgumentParser(
        description=description,
        formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument("args",
                        type=str, metavar="FILE",
                        help="Path to input cif")

    parser.add_argument("-x",
                        type=int, metavar="N", dest="expand_x",
                        help="""Expand N times along x direction""")

    parser.add_argument("-y",
                        type=int, metavar="N", dest="expand_y",
                        help="""Expand N times along y direction""")

    parser.add_argument("-z",
                        type=int, metavar="N", dest="expand_z",
                        help="""Expand N times along z direction""")

    parser.add_argument("-e", "--exclude",
                        type=str, metavar="X", nargs="+", dest="exclude",
                        help="""Exclude these elements from final output""")

    parser.add_argument("-f", "--shift",
                        type=float, metavar="dx dy dz", nargs=3, dest="shift",
                        help="""Applies shift to coordinates, this number comes directly from PLATON (Origin shifted to ... )""")

    parser.add_argument("-s", "--spgr",
                        type=str, metavar="P1", dest="spgr",
                        help="""Apply this symmetry to the new structure (default = P1)""")


    parser.set_defaults(
        expand_x=1,
        expand_y=1,
        expand_z=1,
        exclude=(),
        shift=[],
        spgr="P1"
    )

    options = parser.parse_args()

    cif = options.args
    s = list(read_cif(cif).values())[0]

    excluded = []
    if options.exclude:
        for atom in s.scatterers():
            if atom.element_symbol() in options.exclude:
                excluded.append(atom)

    s = s.expand_to_p1()
    print(f"Expanded to P1 => {s.scatterers().size()} atoms")
    print()

    root, ext = os.path.splitext(cif)

    expand_x = options.expand_x
    expand_y = options.expand_y
    expand_z = options.expand_z

    spgr = options.spgr
    shift = options.shift

    assert expand_x > 0, "N must be bigger than 0"
    assert expand_y > 0, "N must be bigger than 0"
    assert expand_z > 0, "N must be bigger than 0"

    out = root
    for direction, number in enumerate((expand_x, expand_y, expand_z)):
        if number == 1:
            continue
        out += "_{}{}".format(number, 'xyz'[direction])
    if expand_x * expand_y * expand_z == 1:
        if spgr == "P1":
            out += "_P1"
        else:
            out += "_new"
    out += ext

    cell = s.unit_cell().parameters()

    new_cell = (cell[0] * expand_x, cell[1] * expand_y,
                cell[2] * expand_z, cell[3], cell[4], cell[5])

    print("old cell:")
    print(cell)
    print()
    print("new cell:")
    print(new_cell)
    print()


    def expand_cell(scatterers, direction, number):
        for atom in scatterers:
            print(atom.label, number)
            site = list(atom.site)

            coord = site[direction]

            for n in range(number):
                n = float(n)
                new = (coord / number) + (n / number)
                site[direction] = new

                label = atom.label+chr(97+int(n))  # 97 => a
                yield xray.scatterer(label=label, site=site, u=atom.u_iso, occupancy=atom.occupancy)


    if options.exclude:
        scatterers = excluded
        print(">> Excluding these atoms:", ", ".join(options.exclude))
        print()
        for atom in s.scatterers():
            if atom.element_symbol() in options.exclude:
                continue
            else:
                scatterers.append(atom)
    else:
        scatterers = s.scatterers()

    print(f"Starting with {len(scatterers)} scatterers")
    print()

    for direction, number in enumerate((expand_x, expand_y, expand_z)):
        if number == 1:
            continue
        print(" >> Expanding cell along {} by {}".format('xyz'[direction], number))
        scatterers = [
            scatterer for scatterer in expand_cell(scatterers, direction, number)]
        print('New number of scatterers:', len(scatterers))
        print()


    print(f" >> Applying spacegroup {spgr}")
    print()

    sps = crystal.special_position_settings(
        crystal_symmetry=crystal.symmetry(
            unit_cell=new_cell,
            space_group_symbol=spgr),
        min_distance_sym_equiv=0.00001)

    s = xray.structure(
        special_position_settings=sps,
        scatterers=flex.xray_scatterer(scatterers))

    if shift:
        # shift = [-1*number for number in shift]
        print(f">> Applying shift {shift} to all atom sites")
        print()
        s = s.apply_shift(shift)

    if spgr != "P1":
        # shift all atoms to inside unit cell, so asu can be applied
        s = s.sites_mod_positive()

        asu = s.space_group_info().brick().as_string()
        print("Asymmetric unit:")
        # print box_min, "=>", box_max
        print(asu)
        print()
        asu_x, asu_y, asu_z = asu.split(';')
        def fx(x): return eval(asu_x)
        def fy(y): return eval(asu_y)
        def fz(z): return eval(asu_z)

        if '-' in asu:
            print("Did not account for negative values in asymmetric unit. Duplicate atoms cannot be removed (use Kriber)")
            print("")
            print("Remove a,b,c etc from atom labels in cif, then run:")
            print(" > cif2strudat", out)
            print(" > kriber")
            print(" >> reacs global")
            print(" >> wricif")
            print()
        else:
            scatterers = []
            for atom in s.scatterers():
                x, y, z = atom.site

                if fx(x) and fy(y) and fz(z):
                    scatterers.append(atom)

            sps = crystal.special_position_settings(
                crystal_symmetry=crystal.symmetry(
                    unit_cell=new_cell,
                    space_group_symbol=spgr),
                min_distance_sym_equiv=0.00001)

            s = xray.structure(
                special_position_settings=sps,
                scatterers=flex.xray_scatterer(scatterers))

            print(f" >> Removing duplicate atoms, reduced number to {s.scatterers().size()} atoms")
            print()

    s.as_cif_simple(out=open(out, 'w'))
    print(f" >> Wrote file {out}")
    print()

    if (not shift) and (spgr == "P1"):
        print("---")
        print("To find the right symmetry of the expanded unit cell:")
        print()
        print("Run Platon, and then the Addsym routine")
        print("Select NoSubCell in sidebar, then run ADDSYMExact")
        print("Note the space group, and Origin shift")
        print()
        print("Rerun expandcell with --shift X Y Z --spgr SPGR")



import re
import sys


def main():
    pat = re.compile('[a-zA-Z]+')

    if len(sys.argv) == 1:
        print("Usage: fh2topas.py ~/path/to/output.fh [n]")
        sys.exit()

    args = sys.argv[2:]

    moltotal = 1
    if args:
        moltotal = int(args[0])
    molnum = 1

    fin = open(sys.argv[1])

    lines = fin.readlines()

    molnum = 1
    while molnum <= moltotal:
        if moltotal == 1:
            molnum = ''

        number = None
        n = 0
        d = {}
        all_atoms = []

        print('      rigid')

        for line in lines:
            inp = line.split()

            if not inp:
                continue
            if not number:
                number = int(inp[0])
                continue
            n += 1

            print('         z_matrix ', end=' ')

            if n > 0:
                tpe = inp[0]

                atom = f'{tpe}{n}'
                d[str(n)] = atom
                all_atoms.append(atom)

                scatterer = re.findall(pat, atom)[0]
                atom = atom.replace(scatterer, scatterer+str(molnum))
                print(f'{atom:6s}', end=' ')

            if n > 1:
                bonded_with, bond = inp[1:3]
                bonded_with = d[bonded_with]
                scatterer = re.findall(pat, bonded_with)[0]
                bonded_with = bonded_with.replace(
                    scatterer, scatterer+str(molnum))
                print(f'{bonded_with:6s} {bond:7s}', end=' ')
            if n > 2:
                angle_with, angle = inp[3:5]
                angle_with = d[angle_with]
                scatterer = re.findall(pat, angle_with)[0]
                angle_with = angle_with.replace(
                    scatterer, scatterer+str(molnum))
                print(f'{angle_with:6s} {angle:7s}', end=' ')
            if n > 3:
                torsion_with, torsion = inp[5:7]
                torsion_with = d[torsion_with]
                scatterer = re.findall(pat, torsion_with)[0]
                torsion_with = torsion_with.replace(
                    scatterer, scatterer+str(molnum))
                print(f'{torsion_with:6s} {torsion:>7s}', end=' ')

            print()

        print("""
             Rotate_about_axies(@ 0.0 randomize_on_errors,
                                @ 0.0 randomize_on_errors,
                                @ 0.0 randomize_on_errors)
             Translate(         @ 0.0 randomize_on_errors,
                                @ 0.0 randomize_on_errors,
                                @ 0.0 randomize_on_errors)
        """)

        if molnum == '':
            break
        else:
            molnum += 1

    molnum = 1
    while molnum <= moltotal:
        if moltotal == 1:
            molstr = ''
        else:
            molstr = str(molnum)
        print()
        print(f"      prm !occ{molnum} 0.5 min 0.0 max  1.0")
        print(f"      prm !beq{molnum} 3.0 min 1.0 max 10.0")
        print()

        for atom in all_atoms:
            scatterer = re.findall(pat, atom)[0]
            atom = atom.replace(scatterer, scatterer+molstr)
            print(
                f'      site {atom:6s} x 0.0 y 0.0 z 0.0 occ {scatterer:2s} =occ{molnum}; beq =beq{molnum};')
        molnum += 1


if __name__ == '__main__':
    main()



import os
from collections import namedtuple
from tkinter import *
from tkinter.filedialog import *
from tkinter.ttk import *

from future import standard_library

from . import topasdiff

standard_library.install_aliases()


def which(program):
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None


class topasdiffDialog(Tk):

    """Dialog that provide settings window for Topasdiff"""

    def __init__(self, parent, drc='.'):
        super().__init__()

        self.parent = parent

        self.init_vars()

        self.drc = '.'

        self.title("GUI for topasdiff")

        body = Frame(self, padding=(10, 10, 10, 10))
        self.initial_focus = self.body(body)
        body.columnconfigure(0, weight=1)
        body.pack(fill=BOTH, anchor=CENTER, expand=True)

        self.buttonbox()

        self.grab_set()

        self.update()
        self.geometry(self.geometry())

    def init_vars(self):
        self.cif_file = StringVar()
        self.cif_file.set("structure.cif")

        self.fobs_file = StringVar()
        self.fobs_file.set("fobs.hkl")

        self.scale = DoubleVar()
        self.scale.set(1.0)

        self.superflip_path = StringVar()

        self.run_superflip = BooleanVar()
        self.run_superflip.set(True)

    def body(self, master):

        lfcif = Labelframe(master, text="Path to cif file", padding=(10, 10, 10, 10))
        self.e_fname = Entry(
            lfcif, textvariable=self.cif_file)
        self.e_fname.grid(row=11, column=0, columnspan=3, sticky=E+W)
        but_load = Button(lfcif, text="Browse..", width=10, command=self.load_cif_file)
        but_load.grid(row=11, column=4, sticky=E)
        lfcif.grid(row=0, sticky=E+W)
        lfcif.columnconfigure(0, minsize=120)
        lfcif.columnconfigure(0, weight=1)

        lffobs = Labelframe(
            master, text="Path to observed structure factors file", padding=(10, 10, 10, 10))
        self.e_fname = Entry(
            lffobs, textvariable=self.fobs_file)
        self.e_fname.grid(row=21, column=0, columnspan=3, sticky=E+W)
        but_load = Button(lffobs, text="Browse..", width=10, command=self.load_fobs_file)
        but_load.grid(row=21, column=4, sticky=E)
        lffobs.grid(row=1, sticky=E+W)
        lffobs.columnconfigure(0, minsize=120)
        lffobs.columnconfigure(0, weight=1)

        Label(lffobs, text="Scale").grid(row=25, column=0, sticky=W)
        self.e_fname = Entry(
            lffobs, textvariable=self.scale)
        self.e_fname.grid(row=25, column=1, sticky=W)

        if which("superflip"):
            self.superflip_path.set("superflip")
        elif which("superflip.exe"):
            self.superflip_path.set("superflip.exe")
        else:
            lfsf = Labelframe(master, text="Path to superflip executable",
                              padding=(10, 10, 10, 10))
            self.e_fname = Entry(
                lfsf, textvariable=self.superflip_path)
            self.e_fname.grid(row=31, column=0, columnspan=3, sticky=E+W)
            but_load = Button(lfsf, text="Browse..", width=10, command=self.set_superflip_path)
            but_load.grid(row=31, column=4, sticky=E)
            lfsf.grid(row=2, sticky=E+W)
            lfsf.columnconfigure(0, minsize=120)
            lfsf.columnconfigure(0, weight=1)

        # self.c_run_superflip = Checkbutton(lfsf, variable=self.run_superflip, text="Run superflip?")
        # self.c_run_superflip.grid(row=32, column=0, sticky=W)

    def buttonbox(self):
        box = Frame(self)

        w = Button(box, text="Run", width=10, command=self.ok, default=ACTIVE)
        w.pack(side=RIGHT, padx=5, pady=5)
        w = Button(box, text="Exit", width=10, command=self.cancel)
        w.pack(side=RIGHT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

        box.pack(fill=X, anchor=S, expand=False)

    def add_cell(self):
        var = self.init_cell()
        self.vars.append(var)
        self.page(var)

    def ok(self, event=None):
        if not self.validate():
            return

        # self.withdraw()
        self.update_idletasks()

        self.apply()

        # self.destroy()

    def cancel(self, event=None):
        self.destroy()

    def validate(self):
        return 1  # override

    def apply(self):
        os.chdir(self.drc)
        gui_options = {
            "args": self.cif_file.get(),
            "diff": self.fobs_file.get(),
            "superflip_path": self.superflip_path.get(),
            "run_superflip": self.run_superflip.get(),
            "scale": self.scale.get()
        }
        topasdiff.run_script(gui_options=gui_options)

    def load_cif_file(self):

        f = askopenfilename(initialdir=self.drc)
        if f:
            self.cif_file.set(str(f))
            self.drc = os.path.dirname(f)

    def load_fobs_file(self):
        f = askopenfilename(initialdir=self.drc)
        if f:
            self.fobs_file.set(str(f))
            self.drc = os.path.dirname(f)

    def set_superflip_path(self):
        f = askopenfilename(initialdir=self.drc)
        if f:
            self.superflip_path.set(str(f))


def run():
    app = topasdiffDialog(None)
    app.mainloop()


if __name__ == '__main__':
    print(run())



import argparse
import sys

from .blender_mini import *


def print_superflip(sgi, uc, fout, fdiff_file=None):
    """Prints an inflip file that can directly be used with superflip for difference fourier maps

    - Tested and works fine with: EDI, SOD

    sgi: cctbx space_group_info()
    uc : cctbx unit_cell()
    """
    print('title', 'superflip\n', file=fout)

    print('dimension 3', file=fout)
    print('voxel', end=' ', file=fout)
    for p in uc.parameters()[0:3]:
        print(int(((p*4) // 6 + 1) * 6), end=' ', file=fout)
    print(file=fout)
    print('cell', end=' ', file=fout)
    for p in uc.parameters():
        print(p, end=' ', file=fout)
    print('\n', file=fout)

    print('centers', file=fout)
    for cvec in centering_vectors[sgi.type().group().conventional_centring_type_symbol()]:
        print(' '.join(cvec), file=fout)
    print('endcenters\n', file=fout)

    print('symmetry #', sgi.symbol_and_number(), file=fout)
    print('# inverse no', file=fout)

    # number of unique symops, no inverses
    n_smx = sgi.type().group().n_smx()
    # number of primitive symops, includes inverses
    order_p = sgi.type().group().order_p()
    order_z = sgi.type().group().order_z()      # total number of symops

    # this should work going by the assumption that the unique primitive symops are stored first,
    # THEN the inverse symops and then all the symops due to centering.

    for n, symop in enumerate(sgi.type().group()):
        if n == order_p:
            break
        elif n == n_smx:
            print('# inverse yes, please check!', file=fout)
        print(symop, file=fout)

    # Broken, because .inverse() doesn't work, but probably a better approach:
    # for symop in f.space_group_info().type().group().smx():
    #   print >> fout, symop
    # if f.space_group_info().type().group().is_centric():
    #   print >> fout, '# inverse yes'
    #   for symop in f.space_group_info().type().group().smx():
    #       print >> fout, symop.inverse() # inverse does not work?

    print('endsymmetry\n', file=fout)

    print('perform fourier', file=fout)
    print('terminal yes\n', file=fout)

    print('expandedlog yes', file=fout)
    print('outputfile superflip.xplor', file=fout)
    print('outputformat xplor\n', file=fout)

    print('dataformat amplitude phase', file=fout)

    if fdiff_file:
        print('fbegin fdiff.out\n', file=fout)
    else:
        print('fbegin', file=fout)
        print_simple(fcalc, fout, output_phases='cycles')

#       for i,(h,k,l) in enumerate(f.indices()):
#           # structurefactor = abs(f.data()[i])
#           # phase = phase(f.data()[i]
#           print >> fout, "%3d %3d %3d %10.6f %10.3f" % (
#               h,k,l, abs(f.data()[i]), phase(f.data()[i]) / (2*pi) )
        print('endf', file=fout)


def run_script(gui_options=None):
    description = """Notes:
    """

    parser = argparse.ArgumentParser(
        description=description,
        formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument("args",
                        type=str, metavar="FILE",
                        help="Path to input cif")

    parser.add_argument("--diff",
                        type=str, metavar="FILE", dest="diff",
                        help="""Path to file with observed amplitudes to diff with the input cif. Format: h k l F [phase].
                            This uses the indices for the observed reflections. Use the macro 'Out_fobs(fobs.out)' in TOPAS for
                            to output observed structure factors. Scale can be directly copied from topas file or generated automatically.
                            Full command: sfc structure.cif --diff fobs.out""")

    parser.add_argument("-r", "--dmin",
                        type=float, metavar="d_min", dest="dmin",
                        help="maximum resolution")

    parser.add_argument("-t", "--table",
                        type=str, metavar="TABLE", dest="table",
                        help="Choose scattering factor table [x-ray, neutron, electron], default=X-ray")

    parser.set_defaults(
        dmin=None,
        diff=None,
        gui=False,
        superflip_path=None,
        run_superflip=False,
        scale=None,
        table="xray"
    )

    options = parser.parse_args()

    if gui_options:
        for k, v in list(gui_options.items()):
            setattr(options, k, v)

    cif = options.args
    topas_scale = options.scale
    fobs_file = options.diff
    dmin = options.dmin
    table = options.table

    if not cif or not fobs_file:
        print("Error: Supply cif file and use --diff fobs.out to specify file with fobs (hkl + structure factors)")
        sys.exit()

    s = list(read_cif(cif).values())[0]

    uc = s.unit_cell()
    cell = uc.parameters()
    a, b, c, al, be, ga = cell

    sgi = s.space_group().info()
    spgr = str(sgi)

    df = load_hkl(fobs_file, ('fobs',))

    # df = files2df(files)
    if not dmin:
        dmin = calc_dspacing(df, cell, inplace=False).min()

    fcalc = list(calc_structure_factors(cif, dmin=dmin, table=table).values())[0]

    df = df.combine_first(fcalc)

    df = reduce_all(df, cell, spgr)

    # remove NaN reflections (low angle typically)
    df = df[df['fobs'].notnull()]

    if not topas_scale:
        topas_scale = input('Topas scale? >> [auto] ').replace(
            '`', '').replace('@', '').replace('scale', '').strip()

    if topas_scale:
        scale = float(topas_scale)**0.5
        print(f'Fobs scaled by {1/scale} [=sqrt(1/{(float(topas_scale))})]')
    else:
        scale = df['fobs'].sum() / df['fcalc'].sum()
        print(f"No scale given, approximated as {scale} (sum(fobs) / sum(fcal))")

    df['fdiff'] = df['fobs']/scale - df['fcalc']
    df['sfphase'] = df['phases'] / (2*np.pi)

    sel = df['fdiff'] <= 0

    df.loc[sel, 'fdiff'] = abs(df.loc[sel, 'fdiff'])
    df.loc[sel, 'sfphase'] += 0.5

    cols = ('fdiff', 'sfphase')
    write_hkl(df, cols=cols, out='fdiff.out')

    print_superflip(
        sgi, uc, fout=open('sf.inflip', 'w'), fdiff_file='fdiff.out')

    if options.run_superflip:
        import subprocess as sp
        sp.call(f"{options.superflip_path} sf.inflip")


def main(options=None):
    if len(sys.argv) > 1 and sys.argv[1] == "gui":
        from . import topasdiff_gui
        topasdiff_gui.run()
    else:
        run_script()


if __name__ == '__main__':
    main()



import argparse
import io
import os
import sys

from iotbx.cif import CifParserError, reader


def read_cif(f):
    """opens cif and returns cctbx data object"""
    try:
        if isinstance(f, io.IOBase):
            r = reader(file_object=f)
        elif isinstance(f, str):
            r = reader(file_path=f)
        else:
            raise TypeError(f'read_cif: Can not deal with type {type(f)}')
    except CifParserError as e:
        print(e)
        print("Error parsing cif file, check if the data tag does not contain any spaces.")
        sys.exit()
    structures = r.build_crystal_structures()
    for key, val in list(structures.items()):
        print("\nstructure:", key)
        val.show_summary().show_scatterers()
    return structures


def main():
    description = """Notes:
    """

    parser = argparse.ArgumentParser(  # usage=usage,
        description=description,
        formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument("args",
                        type=str, metavar="FILE",
                        help="Path to input cif")

    options = parser.parse_args()

    cif = options.args
    s = list(read_cif(cif).values())[0]

    # s = s.expand_to_p1()

    root, ext = os.path.splitext(cif)
    out = root + "_simple" + ext

    s.as_cif_simple(out=open(out, 'w'))
    print(f" >> Wrote file {out}")


if __name__ == '__main__':
    main()



import argparse
import json
from ast import literal_eval


def parse_str_int_float(item):
    """returns string, int or float either from a list or not"""

    # recursion like a boss
    if isinstance(item, list) and len(item) == 1:
        return parse_str_int_float(item[0])
    elif isinstance(item, list):
        return [parse_str_int_float(i) for i in item]

    try:
        return literal_eval(item)
    except ValueError:
        return item
    except:
        return item


def pprint(d):
    """pretty printing using the json module"""
    print(json.dumps(d, sort_keys=True, indent=4))


def gen_lines(f):
    """return all the lines in every single file"""
    yield from f


def get_lines(lines, options):
    # Indexing_Solutions_With_Zero_Error_2
    # Indexing_Solutions_2

    start = False
    for line in lines:

        if not line:
            continue
        elif line.startswith('Indexing_Solutions'):
            start = not start
        elif ')' in line and start:
            # split on ' and return list of rest, except for the number
            yield line.split("'")[0].replace(')', '').split()
        elif 'Zero' in line and not start:
            options.zero_corr = True
        elif start and line.startswith('}'):
            break


def cosort(*lsts):
    """Takes a few lists and sorts them based on the values of the first list."""
    tmp = list(zip(*lsts))
    tmp.sort()
    return list(zip(*tmp))


def plot_3d(iterator, x_key, y_key, z_key, title="plot", picker=['spgr', 'num']):
    """3d plot: takes 3 keys and an iterator and will plot the values of the keys given"""
    args = [x_key, y_key, z_key] + picker

    xyz_gen = (tuple([d[arg] for arg in args]) for d in iterator)
    x, y, z, spgr, num = list(zip(*xyz_gen))

    import matplotlib.pyplot as plt
    import numpy as np
    from mpl_toolkits.mplot3d import Axes3D

    def onpick(event):
        ind = event.ind

        pckx = np.take(x,    ind)
        pcky = np.take(y,    ind)
        pckz = np.take(z,    ind)
        pckspgr = np.take(spgr, ind)
        pcknum = np.take(num,  ind)

        print()
        for n in range(len(ind)):
            print('idx: {}, {}: {}, {}: {}, {}: {}, spgr: {}, #{}'.format(
                ind[n], x_key, pckx[n], y_key, pcky[n], z_key, pckz[n], pckspgr[n].split('.')[0], pcknum[n]))
        if len(ind) > 5:
            print('number of points:', len(ind))

    def onkeypress(event):
        if event.key == 'x':
            ax.view_init(0, 0)
            plt.draw()
        if event.key == 'y':
            ax.view_init(0, -90)
            plt.draw()
        if event.key == 'z':
            ax.view_init(90, -90)
            plt.draw()

    fig = plt.figure()
    fig.canvas.mpl_connect('pick_event', onpick)
    fig.canvas.mpl_connect('key_press_event', onkeypress)

    ax = fig.add_subplot(111, projection='3d')
    ax.set_xlabel(x_key)
    ax.set_ylabel(y_key)
    ax.set_zlabel(z_key)
    ax.set_title(title)

    ax.scatter(x, y, z, c='r', marker='.', picker=True)
    plt.show()


def plot_2d(iterator, x_key, y_key, picker=['spgr', 'num']):
    """plots given keys in a 2D scatter"""
    args = [x_key, y_key] + picker

    xy_gen = (tuple([d[arg] for arg in args]) for d in iterator)

    x, y, spgr, num = list(zip(*xy_gen))

    import matplotlib.pyplot as plt
    import numpy as np

    def onpick(event):
        ind = event.ind

        pckx = np.take(x,    ind)
        pcky = np.take(y,    ind)
        pckspgr = np.take(spgr, ind)
        pcknum = np.take(num,  ind)

        print()
        for n in range(len(ind)):
            print('idx: {}, {}: {}, {}: {}, spgr: {}, #{}'.format(
                ind[n], x_key, pckx[n], y_key, pcky[n], pckspgr[n].split('.')[0], pcknum[n]))
        if len(ind) > 5:
            print('number of points:', len(ind))

    # col = ax.scatter(x, y, 100*s, c, picker=True) => s,c add color/size to
    # points, maybe nice for best solutions from sflog ??

    fig = plt.figure()
    fig.canvas.mpl_connect('pick_event', onpick)

    ax = fig.add_subplot(111)
    ax.set_xlabel(x_key)
    ax.set_ylabel(y_key)
    ax.set_title(title)

    ax.scatter(x, y, c='r', marker='.', picker=True)
    plt.show()


def plot_1d(iterator, x_key, sort=None, title="Plot", picker=['spgr', 'num']):
    """plots given keys in a 2D scatter"""
    args = [x_key] + picker

    x_gen = (tuple([d[arg] for arg in args]) for d in iterator)

    x, spgr, num = list(zip(*x_gen))

    if sort:
        x, num, spgr = cosort(x, num, spgr)

    import matplotlib.pyplot as plt
    import numpy as np

    def onpick(event):
        ind = event.ind

        pckx = np.take(x,    ind)
        pckspgr = np.take(spgr, ind)
        pcknum = np.take(num,  ind)

        print()
        for n in range(len(ind)):
            print('idx: {}, {}: {}, spgr: {}, #{}'.format(
                ind[n], x_key, pckx[n], pckspgr[n].split('.')[0], pcknum[n]))
        if len(ind) > 5:
            print('number of points:', len(ind))

    # col = ax.scatter(x, y, 100*s, c, picker=True) => s,c add color/size to
    # points, maybe nice for best solutions from sflog ??

    fig = plt.figure()
    fig.canvas.mpl_connect('pick_event', onpick)

    ax = fig.add_subplot(111)
    ax.set_ylabel(x_key)
    ax.set_xlabel('idx')
    ax.set_title(title)

    ax.plot(x, c='r', marker='.', linewidth=0, picker=True)
    plt.show()


def histogram(iterator, x_key, title="Histogram"):
    """plots a histogram of the specified keyword"""
    x = [d[x_key] for d in iterator]

    import matplotlib.mlab as mlab
    import matplotlib.pyplot as plt
    import numpy as np

    fig = plt.figure()
    ax = fig.add_subplot(111)

    n, bins, patches = ax.hist(x, 50, normed=1, facecolor='green', alpha=0.75)

    mu = np.mean(x)    # mean
    sigma = np.std(x)     # standard deviation

    y = mlab.normpdf(bins, mu, sigma)

    ax.plot(bins, y, 'r', linewidth=1)
    ax.set_title(title)

    # print '{:12} - {:12} : {:12}'.format('begin','end','val')
    # for i in range(len(n)):
    #   begin = bins[i]
    #   end   = bins[i+1]
    #   val   = n[i]
    #   print '{:12f} - {:12f} : {:12f}'.format(begin,end,val)

    print("mean:", mu)
    print("sigma:", sigma)

    plt.grid(True)

    plt.show()


def counter(iterator, key):
    """counts the frequency of the value for a given key"""
    freqs = {}
    vals = (d[key] for d in iterator)

    for val in vals:
        if val in freqs:
            freqs[val] += 1
        else:
            freqs[val] = 1

    pprint(freqs)


def table_out(iterator, keywords, out=None):
    """Lists output of specified keywords to stdout (or outfile if an open file object specified)"""
    header = '{:<10}'*len(keywords)
    print(header.format(*keywords), file=out)

    fmt = ''.join([f'{{{keyword}:<10}}' for keyword in keywords])

    for d in iterator:
        print(fmt.format(**d), file=out)


def gen_filter(iterator, args):
    """generator that filters dictionaries based on the input
    args is a list containing the keyword, operator and value"""
    assert len(args) == 3

    kw = args[0]
    op = args[1]
    val = parse_str_int_float(args[2])

    operators = {'gt': '>', 'ge': '>=', 'eq': '==',
                 'ne': '!=', 'le': '<=', 'lt': '<'}

    def conditional(d, kw, val): return eval(f'd[kw] {operators[op]} val')

    for d in iterator:
        if conditional(d, kw, val):
            yield d


def main():
    usage = """superanalyser [KEYWORD ...]
                     [-h] [-v] [-c] [-s N] [-t] [-m] [-i] [-o]
                     [-e KW OP VAL] [-x FILE]"""

    description = """Notes:
- Requires numpy and matplotlib for plotting.
- Keywords/values are case sensitive
- Based on a stripped down version of superanalyser.py, works the same way"""

    parser = argparse.ArgumentParser(usage=usage,
                                     description=description,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument("keywords",
                        type=str, metavar="KEYWORD", nargs='*',
                        help="Keywords for values to analyse. The outcome depends on the number of keywords supplied. With no optional arguments, 1 keyword: histogram; 2 keywords: 2d scatter; 3 keywords: 3d scatter. Run with arguments '-s 0' to see what kind of keywords would be available.")

    parser.add_argument("-x", "--index",
                        action="store", type=str, dest="index_out",
                        help="Output from indexing routine in Topas. Default filename: indexing.ndx")

    parser.add_argument("-c", "--count",
                        action="store_true", dest="count",
                        help="Counts occurances of ARG1 and exits.")

    parser.add_argument("-s", "--show",
                        action="store", type=int, dest="show", metavar='N',
                        help="Generates and prints a single entry with index N and exits.")

    parser.add_argument("-t", "--table",
                        action="store_true", dest="table",
                        help="Prints the values for the given keywords to STDOUT. (default: False)")

    parser.add_argument("-e", "--filter",
                        action="append", dest="filter", nargs=3, metavar='PAR',
                        help="Filters the entries where the value of the keyword does or does not follow the given (in)equality. Order should be 'keyword operator value', where the operator can be gt : >, ge : >=, eq : ==, ne : !=, le : <=, lt : <. Multiple conditionals can be given by chaining them, e.g.: -e spgr eq cmcm -e rsf le 30.")

    parser.add_argument("-m", "--fast",
                        action="store_true", dest="fastmode",
                        help="Turns on fast mode. This will significantly reduce the time taken to parse all the logs, because it only looks for the specified keywords and those needed for --filter. Time taken decreases by up to 500%%.")  # escape % as %%

    parser.add_argument("-i", "--hist",
                        action="store_true", dest="hist",
                        help="Shows a histogram and plots the probability distribution of the first keyword given.")

    parser.add_argument("-o", "--sort",
                        action="store_true", dest="sort",
                        help="Turns on sorting for the 1D plot. Only works for 1D plot. Note: This messes with the indexes, so the idx value can't be used with --show")

    parser.add_argument("-l", "--title",
                        action="store", type=str, dest="title",
                        help="Title for the plot.")

    parser.set_defaults(count=False,
                        table=False,
                        show=None,
                        filter=[],
                        fastmode=False,
                        sort=False,
                        hist=False,
                        debug=False,
                        title='',

                        index_out='indexing.ndx',
                        zero_corr=False)

    options = parser.parse_args()
    args = options.keywords
    print(options, args)

    f = open(options.index_out)

    print(f.name, 'opened')

    lines = gen_lines(f)
    lines = get_lines(lines, options)
    lines = (parse_str_int_float(line) for line in lines)

    if options.zero_corr:
        colnames = ('num', 'spgr', 'status', 'unindexed', 'volume',
                    'zero_corr', 'gof', 'a', 'b', 'c', 'A', 'B', 'C')
    else:
        colnames = ('num', 'spgr', 'status', 'unindexed',
                    'volume', 'gof', 'a', 'b', 'c', 'A', 'B', 'C')

    dicts = (dict(list(zip(colnames, line))) for line in lines)

    # for d in dicts:
    #   pprint(d)
    #   raw_input()

    iterator = dicts

    if options.show is not None:
        for x in range(options.show):
            next(iterator)
        pprint(next(iterator))

    elif options.table:
        table_out(iterator, args)

    elif options.count:
        counter(iterator, args[0])

    elif options.hist:
        histogram(iterator, args[0])

    elif len(args) == 1:
        plot_1d(iterator, args[0])

    elif len(args) == 2:
        plot_2d(iterator, args[0], args[1])

    elif len(args) == 3:
        plot_3d(iterator, args[0], args[1], args[2], title=title)

    elif len(args) > 3:
        table_out(iterator, args)

    else:
        s = 0
        for item in iterator:
            s += 1
        print(s)


if __name__ == '__main__':
    main()



import argparse
import io
import os
import sys

from iotbx.cif import CifParserError, reader


def read_cif(f):
    """opens cif and returns cctbx data object"""
    try:
        if isinstance(f, io.IOBase):
            structures = reader(file_object=f).build_crystal_structures()
        elif isinstance(f, str):
            structures = reader(file_path=f).build_crystal_structures()
        else:
            raise TypeError(f'read_cif: Can not deal with type {type(f)}')
    except CifParserError as e:
        print(e)
        print("Error parsing cif file, check if the data tag does not contain any spaces.")
        sys.exit()
    for key, val in list(structures.items()):
        print("\nstructure:", key)
        val.show_summary().show_scatterers()
    return structures

def main():
    usage = """cif2patterson structure.cif"""

    description = """Notes: Takes any cif file and generated patterson map
    """

    parser = argparse.ArgumentParser(  # usage=usage,
        description=description,
        formatter_class=argparse.RawDescriptionHelpFormatter)


    parser.add_argument("args",
                        type=str, metavar="FILE",
                        help="Path to input cif")


    parser.set_defaults(
        spgr="P1"
    )

    options = parser.parse_args()

    cif = options.args
    s = list(read_cif(cif).values())[0]
    s = s.expand_to_p1()
    print(f"Expanded to P1 => {s.scatterers().size()} atoms")
    print()

    root, ext = os.path.splitext(cif)

    uc = s.unit_cell()

    scatterers = s.scatterers()

    fout1 = open("patterson_dists.txt", 'w')
    fout2 = open("patterson_full.txt", 'w')

    distances_all = []

    verbose = False
    for atom1 in scatterers:
        x1, y1, z1 = atom1.site

        if verbose:
            print()
            atom1.show()
        distances = []
        for atom2 in scatterers:
            if verbose:
                atom2.show()

            x2, y2, z2 = atom2.site

            dx, dy, dz = x1-x2, y1-y2, z1-z2

            dx = dx % 1
            dy = dy % 1
            dz = dz % 1

            length = uc.length((dx, dy, dz))

            distances.append((atom2.label, dx, dy, dz, length))

            if verbose:
                print(f' --> {atom2.label:>4s} {dx:9.5f} {dy:9.5f} {dz:9.5f}  {length:9.5f}')

            # print atom1.label, '-->', atom2.label, '=', uc.length((dx,dy,dz))

        distances_all.extend(distances)

        atom1.show(fout2)
        for label, dx, dy, dz, distance in sorted(distances, key=lambda x: x[-1]):
            print(' --> {:>4s} {:9.5f} {:9.5f} {:9.5f}  {:9.5f}'.format(
                label, dx, dy, dz, distance), file=fout2)
        print(file=fout2)

    print('Wrote file', fout2.name)

    for label, dx, dy, dz, distance in sorted(distances_all, key=lambda x: x[-1]):
        print(f'{distance:9.5f}', file=fout1)

    print('Wrote file', fout1.name)

    fout1.close()
    fout2.close()



import os
import re
import sys


def pbpaste():
    """Mac only: paste text from clipboard"""
    outf = os.popen('pbpaste', 'r')
    content = outf.read()
    outf.close()
    return content


def equals_about(val, compare_to):
    return compare_to * 0.9 <= val <= compare_to*1.1


def main():
    if len(sys.argv) > 1:
        text = open(sys.argv[1]).read()
    else:
        if sys.platform == "darwin":
            text = pbpaste()
        else:
            print('Usage: python restraints_statistics.py [filename]')
            sys.exit()

    oto = 109.5
    tot = 145.0
    to = 1.61

    oto_vals = []
    tot_vals = []
    to_vals = []

    n_lines = 0
    n_oto = 0
    n_tot = 0
    n_to = 0

    for line in re.split('\r|\n', text):
        n_lines += 1

        line = line.strip()

        if line.startswith("'"):
            continue
        if "Restrain" not in line:
            continue
        if "Si" not in line:
            continue

        stuff, restraint, measured, box, weight = line.split(',')

        restraint = float(restraint)
        measured = float(measured.replace('`', '').split("_")[0])

        if equals_about(restraint, oto):
            oto_vals.append(measured)
            n_oto += 1
        elif equals_about(restraint, tot):
            tot_vals.append(measured)
            n_tot += 1
        elif equals_about(restraint, to):
            to_vals.append(measured)
            n_to += 1
        else:
            print(line + ' -- FAIL')

    if not tot_vals:
        tot_vals = [0]
    if not to_vals:
        to_vals = [0]
    if not oto_vals:
        oto_vals = [0]

    print(f"Parsed {n_lines} lines")
    print(f"{n_tot+n_oto+n_to} restraints - tot: {n_tot}, oto: {n_oto}, to: {n_to}")
    print("")
    print('        {:>10s} {:>10s} {:>10s} {:>10s} '.format('restraint', 'min', 'max', 'avg'))
    print(
        f' T-O-T  {tot:10.1f} {min(tot_vals):10.3f} {max(tot_vals):10.3f} {sum(tot_vals)/len(tot_vals):10.3f} ')
    print(
        f' O-T-O  {oto:10.1f} {min(oto_vals):10.3f} {max(oto_vals):10.3f} {sum(oto_vals)/len(oto_vals):10.3f} ')
    print(
        f'   T-O  {to:10.2f} {min(to_vals):10.3f} {max(to_vals):10.3f} {sum(to_vals)/len(to_vals):10.3f} ')


if __name__ == '__main__':
    main()



from itertools import combinations

"""
x. Setting up restraints with topas_restraints.py
    - Go to folder topas_tools/
    - Copy the output from append_bond_lengths for all Si/O atoms to bonds.txt

    Note: The program uses this output to detect with bonds fall within the criteria for zeolites (i.e. bond lengths of 1.61 with some specified tolerance), and uses this connectivity to set up the restraints.

    - Run: >> python topas_restraints.py
        The restraints will be printed to the file restraints.out, these can be copy/pasted directly into the Topas input file.

    Note: There is no filtering of duplicate or symmetry related bonds, so these should be filtered out manually
"""

distdat = {
    ('T',  'O'): (1.610,    0.01),
    ('SI', 'O'): (1.610,    0.01),
    ('AL', 'O'): (1.740,    0.01),
    ('P',  'O'): (1.520,    0.01),
    ('GE', 'O'): (1.750,    0.01),
    ('GA', 'O'): (1.800,    0.01),
    ('B',  'O'): (1.460,    0.01),
    'tot': (109.47,  2.0),
    'oto': (145,  8.0)
}

# location of strings, assuming fixed width
atom1 = [0, 10]
atom2 = [10, 33]
bondl = [33, 41]

organic = False

if not organic:
    # ideal distances/angles
    to_dist = 1.61
    tot_ang = 145.0
    oto_ang = 109.5

    # tolerance for T--O bond detection
    to_tol = 0.2
if organic:
    # ideal distances/angles
    to_dist = 1.70
    tot_ang = 145.0
    oto_ang = 109.5

    # tolerance for X--O bond detection
    to_tol = 0.4


def gen_section(f):
    part = []
    for line in f:
        a1 = line[atom1[0]:atom1[1]]
        a2 = line[atom2[0]:atom2[1]]
        ln = line[bondl[0]:bondl[1]]
        ln = float(ln)
        if a1.split():
            if len(part) > 0:
                part = [item.replace(':', ' ') for item in part]
                print()
                yield part
            part = [a1]
            print(a1, '|', end=' ')

        if to_dist-to_tol < ln < to_dist+to_tol:
            if not a1.split() and len(part) > 1:
                print('           |', end=' ')
            part.append(a2)
            print(a2, '|', ln)
        else:
            pass
    if len(part) > 0:
        part = [item.replace(':', ' ') for item in part]
        print()
        yield part


def main():
    # sigmas
    to_s = 0.01
    tot_s = 10.0
    oto_s = 0.8

    # weights
    to_w = 1/(to_s**2)
    tot_w = 1/(tot_s**2)
    oto_w = 1/(oto_s**2)

    f = open('bonds.txt')
    fout = open('restraints.out', 'w')

    gs = gen_section(f)

    for part in gs:
        main = part[0]
        nbonds = len(part) - 1

        match = 0

        if 'Si' in main or 'Al' in main or 'P' in main:
            match += 1
            if nbonds != 4:
                print(
                    f'*** Warning: More/less than _4_ bonds detected for {main}... bonds = {nbonds}\n')

            for ox in part[1:]:
                print('      Distance_Restrain( {} {} , {}, 0.0, 0.0, {} )'.format(
                    main, ox, to_dist, to_w), file=fout)

            for ox1, ox2 in combinations(part[1:], 2):

                print('      Angle_Restrain( {} {} {} , {}, 0.0, 0.0, {} )'.format(
                    ox1, main, ox2, oto_ang, oto_w), file=fout)

        if 'O' in main:
            match += 1
            if nbonds != 2:
                print(
                    f'*** Warning: More/less than _2_ bonds detected for {main}... bonds = {nbonds}\n')

            for si1, si2 in combinations(part[1:], 2):

                print('      Angle_Restrain( {} {} {} , {}, 0.0, 0.0, {} )'.format(
                    si1, main, si2, tot_ang, tot_w), file=fout)

        if match == 0:
            print(f'*** Non-Si/O detected --> {main}\n')

            for ox in part[1:]:
                print('      Distance_Restrain( {} {} , {}, 0.0, 0.0, {} )'.format(
                    main, ox, to_dist, to_w), file=fout)

            for ox1, ox2 in combinations(part[1:], 2):

                print('      Angle_Restrain( {} {} {} , {}, 0.0, 0.0, {} )'.format(
                    ox1, main, ox2, oto_ang, oto_w), file=fout)


if __name__ == '__main__':
    main()
